import { Injectable } from '@angular/core';
import { 
  HttpClient, 
  HttpHandler, 
  HttpRequest,
  HttpHeaders,
  HttpErrorResponse,
  HttpResponse,
  HttpParams } from '@angular/common/http';

import { Observable } from 'rxjs/Observable';
import { map, retry  } from "rxjs/operators";

import { NotifService } from '../notif';
import { StorageService } from '../storage';
import { UtilsService } from '../utils';
import { CleverQueryService } from './clever-query.service';
import { NG_FOUNDATION_CONFIG } from '../../ng-foundation.config';

export interface HttpOptions{
  body?: any;
  headers?: HttpHeaders | {
      [header: string]: string | string[];
  };
  reportProgress?: boolean;
  observe: 'response';
  params?: HttpParams | {
      [param: string]: string | string[];
  };
  responseType?: 'json';
  withCredentials?: boolean;
}

@Injectable()
export class Agent extends HttpClient{
  
  config: any;
  urls: any;
  headers: HttpHeaders;
  baseDomain: string;
  accessToken: string = null;

  constructor(
    protected nextRequest: HttpHandler,
    protected req: HttpRequest<any>,
    protected notif: NotifService,
    protected store: StorageService,
    protected cleverQuery: CleverQueryService,
    protected _: UtilsService
  ){
    super(nextRequest);
    this.store = this.store.local();

    //initialize request options headers using values from config
    let configHeaders = NG_FOUNDATION_CONFIG['headers'];
    let baseHeaders = new HttpHeaders();
    configHeaders.forEach(function(header){
      baseHeaders.append(header.key, header.value);
    });
    //add jwt token to header
    baseHeaders.append('Authorization', `Bearer ${this.store.get('authToken')}`);

    nextRequest.handle(req.clone({
      headers: baseHeaders
    }));

    
    //set configs
    this.config = NG_FOUNDATION_CONFIG;
    this.urls = NG_FOUNDATION_CONFIG['urls'];
    this.baseDomain = this.urls['base'];

  }
  
  /*
   * high-level features 
   */
  go(){
    this.aPost(this.urls.client.token, {
        security_key: this.config.agent.key,
        security_token: this.config.agent.token
    })
    .map(() => {
      // return Observable.
    });  
  }

  aRequest<R>(method: string, url: string, moreOptions?: HttpOptions): Observable<R>{
    this.onStart();
    return this.request<R>(
      method, url, moreOptions
    )
    .map((response) => {
      return response.body;
    })
    .retry(3);
  }

  aGet<R>(url: string, moreOptions?: HttpOptions): Observable<R>{
    return this.aRequest<R>('GET', url, Object.assign({}, {
      observe: 'response'
    }, moreOptions));
  }

  aPost<R>(url: string, payload: Object, moreOptions?: HttpOptions): Observable<R>{
    return this.aRequest<R>('GET', url, Object.assign({}, {
      observe: 'response',
      body: `payload=${JSON.stringify(payload)}`
    }, moreOptions));
  }

  aPut<R>(url: string, payload: Object, moreOptions?: HttpOptions): Observable<R>{
    return this.aRequest<R>('GET', url, Object.assign({}, {
      observe: 'response',
      body: `payload=${JSON.stringify(payload)}`
    }, moreOptions));
  }
  
  aDelete<R>(url: string, moreOptions?: HttpOptions): Observable<R>{
    return this.aRequest<R>('DELETE', url, Object.assign({}, {
      observe: 'response'
    }, moreOptions));
  }
  
  aPatch<R>(url: string, payload: Object, moreOptions?: HttpOptions): Observable<R>{
    return this.aRequest<R>('PATCH', url, Object.assign({}, {
      observe: 'response',
      body: `payload=${JSON.stringify(payload)}`
    }, moreOptions));
  }

  addAuthToken(url: string): string{
    let authToken = this.store.get('authToken');
    return (url.includes('?')) ? `${url}&auth_token=${authToken}` : `${url}?auth_token=${authToken}`;
  }

  onError (e: HttpErrorResponse){
    let errMsg: string;
    let status = (e.status) ? '' + e.status : '';
    if (e.error instanceof Error) {
        errMsg = `${e.error.name} - ${e.error.message || ''}`;
    } else {
        errMsg = e.error ? e.error : e.toString();
    }
      
    this.notif.toast('HTTP Error - ' + status, errMsg);
    this.notif.complete();
  }
  
  onStart(){
      this.notif.start();
      this.notif.toast('HTTP Action', 'Connecting...');
  }

  onFinish(){
    this.notif.toast('HTTP Action', 'Connecting...Finished.');
    this.notif.complete();
  }
}